import inspect
import json
import uuid
from io import StringIO
from typing import TYPE_CHECKING, Any, Dict, Iterable, List, Union, get_args, get_origin

try:
    from types import UnionType
except ImportError:
    # compatibility with Python 3.9
    UnionType = type(Union[int, str])

import geopandas as gpd
import mercantile
import pandas as pd
import shapely

if TYPE_CHECKING:
    from fused.types import Bbox, TileGDF, TileXYZ


FALSE_CASEFOLDED = "false".casefold()
GEOJSON_CRS = 4326


def _parse_bbox(val: str) -> Union[list, gpd.GeoDataFrame]:
    deserialized = json.loads(val)
    if isinstance(deserialized, list):
        return deserialized
    return gpd.GeoDataFrame.from_features(deserialized, crs=GEOJSON_CRS)


def _parse_to_gdf(val: Any) -> gpd.GeoDataFrame:
    if isinstance(val, gpd.GeoDataFrame):
        return val
    elif isinstance(val, str):
        deserialized = json.loads(val)
        if isinstance(deserialized, list):
            return gpd.GeoDataFrame(
                {}, geometry=[shapely.box(*deserialized)], crs=GEOJSON_CRS
            )
        return gpd.GeoDataFrame.from_features(deserialized, crs=GEOJSON_CRS)
    raise NotImplementedError(f"Not sure how to convert type `{val}` to GeoDataFrame")


def _parse_to_df(val: Any) -> pd.DataFrame:
    if isinstance(val, pd.DataFrame):
        return val
    elif isinstance(val, str):
        return pd.read_json(StringIO(val))
    raise NotImplementedError(f"Not sure how to convert type `{val}` to DataFrame")


def convert_to_tile_gdf(val: Any, add_xyz: bool = True) -> "TileGDF":
    if isinstance(val, str):
        val = _parse_bbox(val)
    if isinstance(val, list):
        val = gpd.GeoDataFrame({}, geometry=[shapely.box(*val)], crs=GEOJSON_CRS)
    if isinstance(val, gpd.GeoDataFrame):
        if add_xyz and not all(c in val.columns for c in ["x", "y", "z"]):
            tile = mercantile.bounding_tile(*val.total_bounds)
            val = val.assign(x=tile.x, y=tile.y, z=tile.z)
            val = val[["x", "y", "z", "geometry"]]
        return val

    raise NotImplementedError(f"Not sure how to convert type `{val}` to TileGDF")


def convert_to_tile_xyz(val: Any) -> "TileXYZ":
    if isinstance(val, str):
        val = _parse_bbox(val)
    if isinstance(val, gpd.GeoDataFrame):
        if all(c in val.columns for c in ["x", "y", "z"]):
            return mercantile.Tile(*val.iloc[0][["x", "y", "z"]])
        else:
            return mercantile.bounding_tile(*val.total_bounds)
    if isinstance(val, list):
        return mercantile.bounding_tile(*val)
    raise NotImplementedError(f"not sure how to convert type `{val}` to TileXYZ")


def convert_to_bbox(val: Any) -> "Bbox":
    if isinstance(val, str):
        val = _parse_bbox(val)
    if isinstance(val, list):
        return shapely.box(*val)
    if isinstance(val, gpd.GeoDataFrame):
        return shapely.box(*val.total_bounds)
    raise NotImplementedError(f"not sure how to convert type `{val}` to Bbox")


def _resolve_annotation(annotation):
    if annotation is inspect._empty:
        return annotation

    origin = get_origin(annotation)
    if origin is not None:
        if origin is Union or origin is UnionType:
            args = get_args(annotation)
            # Handle Optional and (x | None)
            if len(args) == 2:
                if args[0] is type(None):
                    return args[1]
                if args[1] is type(None):
                    return args[0]
            # TODO, handle e.g. Union types. Look into using Pydantic for this. Leave as is for now.
        else:
            return origin
    return annotation


def coerce_arg(val: Any, param: inspect.Parameter) -> Any:
    # Needed due to circular import issue
    from fused.types import Bbox, TileGDF, TileXYZ, ViewportGDF

    # TODO: perhaps use https://docs.python.org/3/library/typing.html#typing.get_type_hints
    annotation = _resolve_annotation(param.annotation)
    if annotation is inspect._empty:
        return val

    # Temporary solution for properly handling non-stdlib annotations
    if isinstance(annotation, str):
        if "GeoDataFrame" in annotation:
            annotation = gpd.GeoDataFrame
        elif "DataFrame" in annotation:
            annotation = pd.DataFrame

    if annotation is int:
        if isinstance(val, str):
            # base 0 will cause the integer to be interpreted as a integer literal similar to how
            # code would be read. Possible improvement to help handle 0x...:
            # https://docs.python.org/3/library/functions.html#int
            # Only strings should be passed in to `int` if the base is passed too. Integers and so
            # on will not be converted correctly.
            return int(val, base=10)
        return int(val)
    if annotation is float:
        return float(val)
    if annotation is bool:
        if isinstance(val, str) and val.casefold() == FALSE_CASEFOLDED:
            return False
        return bool(val)
    if annotation in [list, dict, List, Dict, Iterable] and isinstance(val, str):
        return json.loads(val)
    if annotation is tuple:
        return tuple(json.loads(val))
    if annotation is uuid.UUID:
        return uuid.UUID(val)
    if annotation in [TileGDF, ViewportGDF]:
        return convert_to_tile_gdf(val, add_xyz=annotation is TileGDF)
    if annotation is TileXYZ:
        return convert_to_tile_xyz(val)
    if annotation is Bbox:
        return convert_to_bbox(val)
    if annotation is gpd.GeoDataFrame:
        # load as geojson
        return _parse_to_gdf(val)
    if annotation is pd.DataFrame:
        # load as json
        return _parse_to_df(val)

    # Unknown, fall back to not doing anything with it
    return val



import json
from tempfile import NamedTemporaryFile


def deserialize_tiff(content: bytes):
    import rasterio
    import rioxarray
    import xarray as xr

    with NamedTemporaryFile(prefix="udf_result", suffix=".tiff") as ntmp:
        with open(ntmp.name, "wb") as f:
            f.write(content)
        with rasterio.open(ntmp.name) as dataset:
            rda = rioxarray.open_rasterio(
                dataset,
                masked=True,
            )
            metadata = dataset.tags(ns="fused")
        orig_type = metadata.get("orig_type")
        if orig_type == "numpy.ndarray":
            data = rda.values
            if len(json.loads(metadata["shape"])) < rda.ndim:
                data = data.squeeze()
            if "bounds" in metadata:
                bounds = json.loads(metadata["bounds"])
                data = (data, bounds)
        elif orig_type == "xarray.DataArray":
            data = rda
        else:
            key = metadata.get("key") or "image"
            data = xr.Dataset({key: rda})
    return data



from __future__ import annotations

import warnings
from pathlib import Path
from typing import (
    Any,
    BinaryIO,
    Dict,
    Iterable,
    Literal,
    Optional,
    Sequence,
    Union,
    overload,
)

from pydantic import Field, PrivateAttr, RootModel
from typing_extensions import Annotated

from fused.models.schema import Schema
from fused.models.udf.base_udf import (
    METADATA_FUSED_ID,
    METADATA_FUSED_SLUG,
    BaseUdf,
    UdfType,
)
from fused.warnings import FusedUdfWarning

from .._inplace import _maybe_inplace


class GeoPandasUdfV2(BaseUdf):
    """A user-defined function that operates on [`geopandas.GeoDataFrame`s][geopandas.GeoDataFrame]."""

    type: Literal[UdfType.GEOPANDAS_V2] = UdfType.GEOPANDAS_V2
    table_schema: Optional[Schema] = None
    """The [`Schema`][fused.models.Schema] describing the output of this UDF.
    """

    entrypoint: str
    """Name of the function within the code to invoke."""

    parameters: Dict[str, Any] = Field(default_factory=dict)
    """Parameters to pass into the entrypoint."""

    _parameter_list: Optional[Sequence[str]] = PrivateAttr(None)
    original_headers: Optional[str] = None

    _nested_callable = PrivateAttr(None)  # TODO : Find out type

    def set_parameters(
        self,
        parameters: Dict[str, Any],
        replace_parameters: bool = False,
        inplace: bool = False,
    ) -> "GeoPandasUdfV2":
        """Set the parameters on this UDF.

        Args:
            parameters: The new parameters dictionary.
            replace_parameters: If True, unset any parameters not in the parameters argument. Defaults to False.
            inplace: If True, modify this object. If False, return a new object. Defaults to True.
        """
        ret = _maybe_inplace(self, inplace)
        new_parameters = (
            parameters
            if replace_parameters
            else {
                **ret.parameters,
                **parameters,
            }
        )
        ret.parameters = new_parameters
        return ret

    def eval_schema(self, inplace: bool = False) -> "GeoPandasUdfV2":
        """Reload the schema saved in the code of the UDF.

        Note that this will evaluate the UDF function.

        Args:
            inplace: If True, update this UDF object. Otherwise return a new UDF object (default).
        """
        from fused._udf.execute_v2 import execute_for_decorator

        new_udf = execute_for_decorator(self)
        assert isinstance(
            new_udf, GeoPandasUdfV2
        ), f"UDF has unexpected type: {type(new_udf)}"
        ret = _maybe_inplace(self, inplace)
        ret.table_schema = new_udf.table_schema
        ret._parameter_list = new_udf._parameter_list
        return ret

    def run_local(
        self,
        sample: Any | None = ...,
        *,
        inplace: bool = False,
        validate_output: bool = False,
        validate_imports: Optional[bool] = None,
        **kwargs,
    ) -> "UdfEvaluationResult":  # ruff: noqa: F821
        """Evaluate this UDF against a sample.

        Args:
            sample: Sample (from `get_sample`) to execute against.
            inplace: If True, update this UDF object with schema information. (default)
        """
        from fused._udf.execute_v2 import execute_against_sample
        from fused.models.udf._eval_result import (  # ruff: noqa: F401
            UdfEvaluationResult,
        )

        ret = _maybe_inplace(self, inplace)
        sample_list = [] if sample is Ellipsis else [sample]
        return execute_against_sample(
            udf=ret,
            input=sample_list,
            validate_output=validate_output,
            validate_imports=validate_imports,
            **kwargs,
        )

    def to_file(self, where: Union[str, Path, BinaryIO], *, overwrite: bool = False):
        """Write the UDF to disk or the specified file-like object.

        The UDF will be written as a Zip file.

        Args:
            where: A path to a file or a file-like object.

        Keyword Args:
            overwrite: If true, overwriting is allowed.
        """
        job = self()
        job.export(where, how="zip", overwrite=overwrite)

    def to_directory(self, where: Union[str, Path], *, overwrite: bool = False):
        """Write the UDF to disk as a directory (folder).

        Args:
            where: A path to a directory.

        Keyword Args:
            overwrite: If true, overwriting is allowed.
        """
        job = self()
        job.export(where, how="local", overwrite=overwrite)

    # List of data input is passed - run that
    @overload
    def __call__(self, *, arg_list: Iterable[Any], **kwargs) -> UdfJobStepConfig:
        ...

    # Nothing is passed - run the UDF once
    @overload
    def __call__(self, *, arg_list: None = None, **kwargs) -> UdfJobStepConfig:
        ...

    def __call__(
        self, *, arg_list: Optional[Iterable[Any]] = None, **kwargs
    ) -> Union[UdfJobStepConfig,]:
        """Create a job from this UDF.

        Args:
            arg_list: A list of records to pass in to the UDF as input.
        """
        # cyclic dependency if imported at top-level
        from fused.models.api.job import UdfJobStepConfig

        with_params = self.model_copy()
        # TODO: Consider using with_parameters here, and validating that "context" and other reserved parameter names are not being passed.
        new_parameters = {**kwargs}
        if new_parameters:
            with_params.parameters = new_parameters

        if arg_list is not None and not len(arg_list):
            warnings.warn(
                FusedUdfWarning(
                    "An empty `arg_list` was passed in, no calls to the UDF will be made."
                )
            )

        return UdfJobStepConfig(
            udf=with_params,
            input=arg_list,
        )


EMPTY_UDF = GeoPandasUdfV2(
    name="EMPTY_UDF", code="", entrypoint="", table_schema=Schema(fields=[])
)

AnyBaseUdf = Annotated[GeoPandasUdfV2, Field(..., discriminator="type")]


class RootAnyBaseUdf(RootModel[AnyBaseUdf]):
    pass


def load_udf_from_response_data(data: dict) -> RootAnyBaseUdf:
    """Return a UDF from an HTTP response body, adding in metadata if necessary"""

    udf = RootAnyBaseUdf.model_validate_json(data["udf_body"]).root
    # Restore metadata fields if they were not already present
    if not udf._get_metadata_safe(METADATA_FUSED_ID) and "id" in data:
        udf._set_metadata_safe(METADATA_FUSED_ID, data["id"])
    if not udf._get_metadata_safe(METADATA_FUSED_SLUG) and "slug" in data:
        udf._set_metadata_safe(METADATA_FUSED_SLUG, data["slug"])
    return udf



import base64
import functools
import hashlib
import inspect
import os
import pickle
import types
from typing import Any, Awaitable, Callable, Optional, TypeVar

from loguru import logger

from fused._options import options as OPTIONS


def _serialize_code(code: types.CodeType) -> bytes:
    func_code_bytes = base64.b64encode(code.co_code)

    # Introduce variable values into hash.
    func_const_values = [
        # Embedded lambdas must be serialized, same as the overall function
        (
            _serialize_code(const)
            if isinstance(const, types.CodeType)
            else str(const).encode("utf-8")
        )
        for const in code.co_consts
    ]
    # Names must be serialized as otherwise referenced names (min vs max) would not
    # affect cache key.
    func_name_values = [name.encode("utf-8") for name in code.co_names]

    return b"".join(
        [
            func_code_bytes,
            *func_const_values,
            *func_name_values,
        ]
    )


def _serialize_function_defaults(func: Callable) -> bytes:
    parts = []

    if func.__defaults__ is not None:
        for d in func.__defaults__:
            parts.append(_hashify(d).encode("utf-8"))

    if func.__kwdefaults__ is not None:
        for key, value in func.__kwdefaults__.items():
            parts.append(f"{key}{_hashify(value)}".encode("utf-8"))

    return b"".join(parts)


def _hashify(func) -> str:
    hash_object = hashlib.sha256()
    try:
        if hasattr(func, "__fused_cached_fn"):
            return _hashify(func.__fused_cached_fn)
        elif callable(func):
            hash_object.update(_serialize_code(func.__code__))
            # Caution! The defaults and args do not go into the same part of the cache key!
            hash_object.update(_serialize_function_defaults(func))
        else:
            hash_object.update(str(func).encode("utf-8"))
        return hash_object.hexdigest()
    except Exception as e:
        logger.warning(f"Error Hashing {e}")
        return ""


def _cache(
    func: Callable,
    *args,
    reset: bool = False,
    path: str = "tmp",
    retry: bool = True,
    **kwargs,
):
    path = path.strip("/")

    # Cache directory
    # TODO: consider udf name in path once available from Fused global context
    path = OPTIONS.cache_directory / path
    path.mkdir(parents=True, exist_ok=True)

    # Pop reserved `_cache_id`kwarg
    _cache_id = kwargs.pop("_cache_id", None)

    try:
        # TODO: ignore `_`

        # 1. Hashify function
        id = _hashify(func)

        # 2. Hashify args
        for v in args:
            id += "_" + _hashify(v)

        # 3. Hashify kwargs
        for k in kwargs:
            id += k + _hashify(kwargs[k])

        # 4. Hashify _cache_id
        id += _hashify(_cache_id)

        # 5. Hashify composite id
        id = _hashify(id)

        path_file = path / f"data_{id}"

        if not os.path.exists(path_file) or reset:
            with open(path_file, "wb") as f:
                data = func(*args, **kwargs)
                pickle.dump(data, f, protocol=pickle.HIGHEST_PROTOCOL)
                return data
        else:
            try:
                with open(path_file, "rb") as f:
                    data = pickle.load(f)

                    print(f"{func.__qualname__} was cached")
                    return data
            except Exception:
                logger.debug("Cache not found, retrying.")
                if retry:
                    with open(path_file, "wb") as f:
                        data = func(*args, **kwargs)
                        pickle.dump(data, f, protocol=pickle.HIGHEST_PROTOCOL)
                        return data
                else:
                    return None
    except Exception as e:
        logger.info(f"Error Caching {e}")
        raise e


async def _cache_async(
    func: Callable,
    *args,
    reset: bool = False,
    path: str = "tmp",
    retry: bool = True,
    **kwargs,
):
    path = path.strip("/")

    # Cache directory
    # TODO: consider udf name in path once available from Fused global context
    path = OPTIONS.cache_directory / path
    path.mkdir(parents=True, exist_ok=True)

    # Pop reserved `_cache_id`kwarg
    _cache_id = kwargs.pop("_cache_id", None)

    try:
        # TODO: ignore `_`

        # 1. Hashify function
        id = _hashify(func)

        # 2. Hashify args
        for v in args:
            id += "_" + _hashify(v)

        # 3. Hashify kwargs
        for k in kwargs:
            id += k + _hashify(kwargs[k])

        # 4. Hashify _cache_id
        id += _hashify(_cache_id)

        # 5. Hashify composite id
        id = _hashify(id)

        path_file = path / f"data_{id}"

        if not os.path.exists(path_file) or reset:
            with open(path_file, "wb") as f:
                data = await func(*args, **kwargs)
                pickle.dump(data, f, protocol=pickle.HIGHEST_PROTOCOL)
                return data
        else:
            try:
                with open(path_file, "rb") as f:
                    data = pickle.load(f)

                    print(f"{func.__qualname__} was cached")
                    return data
            except Exception:
                logger.debug("Cache not found, retrying.")
                if retry:
                    with open(path_file, "wb") as f:
                        data = await func(*args, **kwargs)
                        pickle.dump(data, f, protocol=pickle.HIGHEST_PROTOCOL)
                        return data
                else:
                    return None
    except Exception as e:
        logger.info(f"Error Caching {e}")
        raise e


def _cache_internal(func, **decorator_kwargs):
    def decorator(func):
        _path = decorator_kwargs.get("path", "tmp")

        if inspect.iscoroutinefunction(func):

            @functools.wraps(func)
            async def wrapper_async(*args, **kwargs):
                return await _cache_async(func, *args, path=_path, **kwargs)

            wrapper_async.__fused_cached_fn = func

            return wrapper_async
        else:

            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                return _cache(func, *args, path=_path, **kwargs)

            wrapper.__fused_cached_fn = func

            return wrapper

    if callable(func):  # w/o args
        return decorator(func)
    else:  # w/ args
        return decorator


def cache(
    func: Optional[Callable[..., Any]] = None, **kwargs: Any
) -> Callable[..., Any]:
    """Decorator to cache the return value of a function.

    This function serves as a decorator that can be applied to any function
    to cache its return values. The cache behavior can be customized through
    keyword arguments.

    Args:
        func (Callable, optional): The function to be decorated. If None, this
            returns a partial decorator with the passed keyword arguments.
        **kwargs: Arbitrary keyword arguments that are passed to the internal
            caching mechanism. These could specify cache size, expiration time,
            and other cache-related settings.

    Returns:
        Callable: A decorator that, when applied to a function, caches its
        return values according to the specified keyword arguments.

    Examples:

        Use the `@cache` decorator to cache the return value of a function in a custom path.

        ```py
        @cache(path="/tmp/custom_path/")
        def expensive_function():
            # Function implementation goes here
            return result
        ```

        If the output of a cached function changes, for example if remote data is modified,
        it can be reset by running the function with the `reset` keyword argument. Afterward,
        the argument can be cleared.

        ```py
        @cache(path="/tmp/custom_path/", reset=True)
        def expensive_function():
            # Function implementation goes here
            return result
        ```
    """
    return _cache_internal(func=func, **kwargs)


T = TypeVar("T")


def cache_call(func: Callable[..., T], *args: Any, **kwargs: Any) -> T:
    """Directly calls a function with caching.

    This function directly calls the provided function with the given arguments
    and keyword arguments, caching its return value. The cache used depends on
    the implementation of the `_cache` function.

    Args:
        func (Callable): The function to call and cache its result.
        *args: Positional arguments to pass to the function.
        **kwargs: Keyword arguments to pass to the function.

    Returns:
        The cached return value of the function.

    Raises:
        Exception: Propagates any exception raised by the function being called
        or the caching mechanism.
    """
    return _cache(func, *args, **kwargs)


async def cache_call_async(
    func: Callable[..., Awaitable[T]], *args: Any, **kwargs: Any
) -> T:
    """Asynchronously calls a function with caching.

    Similar to `cache_call`, but for asynchronous functions. This function
    awaits the provided async function, caches its return value, and then
    returns it. The specifics of the caching mechanism depend on the
    implementation of `_cache_async`.

    Args:
        func (Callable): The asynchronous function to call and cache its result.
        *args: Positional arguments to pass to the async function.
        **kwargs: Keyword arguments to pass to the async function.

    Returns:
        The cached return value of the async function.

    Raises:
        Exception: Propagates any exception raised by the async function being
        called or the caching mechanism.

    Examples:
        async def fetch_data(param):
            # Async function implementation goes here
            return data

        # Usage
        result = await cache_call_async(fetch_data, 'example_param')
    """
    return await _cache_async(func, *args, **kwargs)
